import numpy as np

from .shape import SAMPLERATE


class Tone(object):
    """A stereo signal  of given frequency, duration and panning position"""

    def __init__(self, frequency, duration, deviation=0):
        self.frequency = frequency
        self.duration = duration
        self.deviation = deviation

    @property
    def values(self):
        return self.__pan(self.__tone())

    def __tone(self) -> np.ndarray:
        """Returns a sine wave of given frequency and duration"""
        t = np.linspace(0, self.duration, int(SAMPLERATE * self.duration), False)
        return np.sin(2 * np.pi * self.frequency * t)

    def __pan(self, values: np.ndarray) -> np.ndarray:
        """Multiplies values by amplitude and transforms mono signal
        into a set of two   channels with effect
        of audible horizontal deviation from the central point
        """

        left = (
            np.sqrt(2)
            / 2.0
            * (np.cos(self.deviation) - np.sin(self.deviation))
            * values
        )
        right = (
            np.sqrt(2)
            / 2.0
            * (np.cos(self.deviation) + np.sin(self.deviation))
            * values
        )
        return np.vstack((left, right))



import numpy as np


def singleton(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


@singleton
class Space(object):
    """A  space defined by its dimensions, the class implements singleton pattern
    - there can be only one Space object during  the  session,
    all parameters are read only and cannot be modified"""

    class Dimension:
        """An inner class stores  the range of possible  values and encapsulates a method for   rescaling input values into that range"""

        def __init__(self, limits):
            self.__limits = limits

        @property
        def limits(self):
            return self.__limits

        def bind(self, values):
            return self.__rescale(values, *self.limits)

        def __rescale(self, values: np.ndarray, min: int, max: int) -> np.ndarray:
            """Rescales the passed values into the  range between min and max."""
            values_max, values_min = np.max(values), np.min(values)
            # in case of a single value
            if values_max == values_min:
                if values_max > len(values):
                    values[:] = len(values)
                elif values_min < 0:
                    values[:] = 0
                values_min, values_max = 0, len(values)
            return (values - values_min) / (values_max - values_min) * (max - min) + min

    def __init__(
        self,
        x_limits=(-45, 45),  # degrees
        y_limits=(220, 7040),  # A of small and fifth octave
    ):
        self.__x = self.Dimension(x_limits)
        self.__y = self.Dimension(y_limits)

    @property
    def x(self):
        return self.__x

    @property
    def y(self):
        return self.__y

    def __repr__(self):
        return f"Space ({self.__x.limits} degrees, {self.__y.limits} Hz)"



import numpy as np


def smoothen(values: np.ndarray, window_size: int = 3) -> np.ndarray:
    """Applies the simple moving averages algorythm with passed window size to the values
    starting with values[0] + window_size element.
    The first window_size elements are replaced by values' minimum.
    """

    shift = np.array([float(np.min(values))] * window_size)
    return np.hstack(
        (
            shift,
            np.array(
                [
                    np.sum(values[i : i + window_size]) / window_size
                    for i in range(len(values[:-window_size]))
                ]
            ),
        )
    )



import numpy as np

from .shape import Shape
from .preprocessing import smoothen


def arg_manager(class_):
    def set_params(*args, **kwargs):
        defaults = dict(
            {
                "bins": 100,
                "density": False,
                "smooth": False,
                "window_size": 3,
                "mode": "fast",
            }
        )
        for arg in defaults.keys():
            if arg in kwargs.keys():
                defaults[arg] = kwargs[arg]
        return class_(*args, **defaults)

    return set_params


@arg_manager
class Histogram(Shape):
    """A decorated constructor function takes and then safely routes
    any set of optional parameters from  default dictionary,
    including some parameters of numpy.histogram(),
    smoothen() and constructor of the parent class itself:
    bins: int,  number of bins in histogram, default is 100,
    density: bool, see the density parameter of numpy.histogram()
        default is False,
    smooth: bool,  whether to apply simple moving averages algorythm
        to the values or not, default is False,
    window_size: int, window size for sma, makes sence only if smooth=True.
    mode: {'fast', 'slow'}, corresponds to the length of segment
        of the Shape in seconds, 0.01 and 0.05 accordingly
    """

    def __init__(self, values, **kwargs):
        if kwargs is None:
            raise ValueError("Default settings not found")
        bins, density, smooth, window_size, mode = kwargs.values()
        x = np.arange(bins)
        y = np.histogram(values, bins, density=density)[0]
        if smooth:
            y = smoothen(y, window_size)
        super().__init__(x, y, mode)



SAMPLERATE = 44100

import numpy as np
import sounddevice as sd
import soundfile as sf
from datetime import datetime

from .space import Space
from .tone import Tone


class Shape(object):
    """An object defined by x and y coordinates
    scaled to fit  the borders of space attribute.

    x and y are single-dimensional arrays

    Can contain an embedded object of the same type
    with implemented recursive access to the values

    methods defined here:
    to_device(): sends the values to sound device
    to_wav(): creates a wav-file with values
    switch(mode: {'fast', 'slow'}): switches the mode
    add(x: np.ndarray, y: np.ndarray): adds embedded object of the same type

    Note: When using Shape object in context of Jupyter notebook
    it is strongly recommended to pass its values directly to the
    IPython.display.Audio() method, which provides
    play/pause control,  so you can easily replay the sound
    without executing the code again.

    Example:

    from IPython.display import Audio
    sr = 1000
    f = 7
    samples = 1000
    x = np.arange(samples)
    y = np.sin(2 * np.pi * f * x / sr)
    sinusoid  = Shape(x, y)
    Audio(sinusoid.values, rate=44100)

    in any other cases you should use to_device() method instead
    and replace just the last line like this:

    sinusoid.to_device()

    Note: To create a histogram, you can use the  Shape object with proper x and y,
    but there is an option   to use an instance  of Histogram class,
    which inherits everything from Shape
    but its constructor takes  a distribution itself and a variaty of
    optional keyword arguments as well.

    Example:

    from scipy.stats import norm
    normal_distribution = norm.rvs(size=10000)
    histogram = Histogram(normal_distribution, bins=100, density=False, smooth=False, window_size=3, mode='fast')
    # and either
    Audio(histogram.values, rate=44100)
    # or just
    histogram.to_device()

    Note: Because any space is square-shaped,
    it can  produce some unexpected results when using embedded shapes -
    a line which is single and neither horizontal nor  vertical, is diagonal.

    """

    def __init__(self, x=np.array([0]), y=np.array([0]), mode="fast"):
        if len(x) != len(y):
            raise ValueError("Arguments x and y must be of the same length")
        self.space = Space()
        self.switch(mode)
        self.__x = x
        self.__y = y
        self.__values = None

    @property
    def x(self):
        return np.radians(self.space.x.bind(self.__x))

    @x.setter
    def x(self, values):
        self.__x = values

    @property
    def y(self):
        return self.space.y.bind(self.__y)

    @y.setter
    def y(self, values):
        self.__y = values

    @property
    def values(self):
        """Returns values shaped like (n_channels, n_samples)"""
        values = np.array([[], []])
        for x_value, y_value in zip(self.x, self.y):
            values = np.hstack(
                (
                    values,
                    Tone(
                        frequency=y_value,
                        duration=self.segment_duration,
                        deviation=x_value,
                    ).values,
                )
            )
        values = np.nan_to_num(values)
        if self.__values is not None:
            return np.hstack((values, self.__values.values))
        return values

    def switch(self, mode: {"fast", "slow"}) -> None:
        """Sets either 0.01 (fast mode) or 0.05 (slow mode)"""
        modes = {"fast": 0.01, "slow": 0.05}
        if mode not in modes.keys():
            raise ValueError(f"{mode} is not a valid value for the argument")
        self.segment_duration = modes[mode]

    def add(self, x, y):
        if self.__values == None:
            self.__values = Shape(x, y)
            self.__values.segment_duration = self.segment_duration
        else:
            self.__values.add(x, y)

    def to_device(self):
        sd.play(self.values.T, samplerate=SAMPLERATE, blocking=False)
        sd.wait()

    def to_wav(self):
        now = datetime.now()
        time_format = "%Y-%m-%d_%H-%M-%S"
        filename = f"{now:{time_format}}.wav"
        sf.write(filename, self.values.T, SAMPLERATE)
